import sqlite3
import os
import shutil
import re
import random
import json
import zipfile
import tempfile
from copy import deepcopy
from collections import OrderedDict


MW_DOMAINS = ['restaurant', 'hotel', 'attraction', 'train', 'taxi', 'hospital', 'police']


def load_databases(zipf):
    dbs = {}
    sql_dbs = {'attraction', 'hotel', 'restaurant', 'train'}
    for domain in MW_DOMAINS:
        if domain in sql_dbs:
            db = 'db/{}-dbase.db'.format(domain)
            with tempfile.NamedTemporaryFile('rb+') as dbf:
                shutil.copyfileobj(zipf.open(db), dbf)
                dbf.flush()
                fileconn = sqlite3.connect(dbf.name)
                conn = sqlite3.connect(':memory:')
                fileconn.backup(conn)

            def dict_factory(cursor, row):
                d = {}
                for idx, col in enumerate(cursor.description):
                    d[col[0]] = row[idx]
                return d

            conn.row_factory = dict_factory
            c = conn.cursor()
            dbs[domain] = c
        else:
            db = 'db/{}_db.json'.format(domain)
            dbs[domain] = json.load(zipf.open(db))
    return dbs


class Database:
    def __init__(self, zipf, seed=42):
        self.path = zipf.filename
        self.dbs = load_databases(zipf)
        self.ignore_values = ['not mentioned', 'dont care', 'don\'t care', 'dontcare', 'do n\'t care', 'none']
        self.rng = random.Random(seed)

    price_re = re.compile(r'\d+\.\d+')

    @staticmethod
    def translate_to_db_col(s):
        if s == 'leave at':
            return 'leaveAt'
        elif s == 'arrive by':
            return 'arriveBy'
        elif s == 'price range':
            return 'pricerange'
        else:
            return s

    def domain_not_empty(self, domain_bs):
        return any(len(val) > 0 and val not in self.ignore_values for val in domain_bs.values())

    @staticmethod
    def map_database_key(key):
        if key == 'trainID':
            key = 'id'
        key = ''.join([' '+i.lower() if i.isupper()
                       else i for i in key]).lstrip(' ')
        key = key.replace('_', ' ')
        if key == 'pricerange':
            key = 'price range'
        if key == 'taxi phone' or key == 'phone':
            key = 'phone'
        if key == 'taxi colors':
            key = 'color'
        if key == 'taxi types':
            key = 'brand'
        if key == 'ref':
            key = 'reference'
        if key == 'leaveAt':
            key = 'leave at'
        if key == 'arriveBy':
            key = 'arrive by'
        if key == 'entrance fee':
            key = 'fee'
        return key

    @staticmethod
    def map_query_value(value):
        if value == 'concert hall':
            value = 'concerthall'
        if value == 'night club':
            value = 'nightclub'
        return value

    @staticmethod
    def capitalize(val):
        def _mk(v):
            i, v = v
            if i == 0 or v not in {'the', 'an', 'a', 'of', 'in', 'for', 'as', 'these', 'at', 'up', 'on', 'and', 'or'}:
                return v[:1].upper() + v[1:]
            else:
                return v
        return ' '.join(map(_mk, enumerate(val.split())))

    @staticmethod
    def map_database_row(domain, row, query):
        results = dict()
        for k, val in row.items():
            k2 = Database.map_database_key(k)
            if k == 'location':
                continue
            elif k == 'post code' or k == 'postcode':
                val = val.upper()
            elif k == 'name':
                val = Database.capitalize(val)
            elif k == 'type' and val == 'concerthall':
                val = 'concert hall'
            elif k == 'price' and domain == 'hotel' and isinstance(val, dict):
                val = val.get('single', val.get('double', next(iter(val.values()))))
                val = f'{val} pounds'
            if k2 == 'people':
                # BUG in MW2.0
                val = val.lstrip('`')
            results[k2] = val
        if 'color' in results and 'brand' in results:
            results['car'] = f"{results['color']} {results['brand']}"
        if domain == 'train' and 'price' in row and 'people' in query:
            people = int(query['people'])

            def multiply_people(m):
                price = float(m.group(0))
                price *= people
                return format(price, '.2f')
            if people != 1:
                results['price'] = Database.price_re.sub(multiply_people, row['price'])
        return results

    def query_domain(self, domain, query):
        # Handle special domains not in sqlite databases
        # NOTE: this is not a part of multiwoz repo
        # Taken from convlab
        if domain == 'taxi':
            return [{'color': self.rng.choice(self.dbs[domain]['taxi_colors']),
                     'brand': self.rng.choice(self.dbs[domain]['taxi_types']),
                     'phone': ''.join([str(random.randint(1, 9)) for _ in range(11)])}]
        if domain == 'police':
            return deepcopy(self.dbs['police'])
        if domain == 'hospital':
            department = None
            for key, val in query:
                if key == 'department':
                    department = val
            if not department:
                return deepcopy(self.dbs['hospital'])
            else:
                return [deepcopy(x) for x in self.dbs['hospital']
                        if x['department'].lower() == department.strip().lower()]

        sql_query = "select * from {}".format(domain)

        flag = True
        for key, val in query:
            if val == "" or val in self.ignore_values:
                pass
            else:
                if flag:
                    sql_query += " where "
                    val2 = val.replace("'", "''")
                    # change query for trains
                    if key == 'leaveAt':
                        sql_query += r" " + key + " > " + r"'" + val2 + r"'"
                    elif key == 'arriveBy':
                        sql_query += r" " + key + " < " + r"'" + val2 + r"'"
                    else:
                        sql_query += r" " + key + "=" + r"'" + val2 + r"'"
                    flag = False
                else:
                    val2 = val.replace("'", "''")
                    if key == 'leaveAt':
                        sql_query += r" and " + key + " > " + r"'" + val2 + r"'"
                    elif key == 'arriveBy':
                        sql_query += r" and " + key + " < " + r"'" + val2 + r"'"
                    else:
                        sql_query += r" and " + key + "=" + r"'" + val2 + r"'"

        result = self.dbs[domain].execute(sql_query).fetchall()
        return result

    def __call__(self, belief, return_results=False):
        all_results = OrderedDict()
        for domain, domain_bs in belief.items():
            blocked_slots = {'people', 'booked', 'stay'}
            if domain != 'train' and domain != 'bus':
                blocked_slots.add('day')
                blocked_slots.add('time')
            query = [(Database.translate_to_db_col(slot), Database.map_query_value(val))
                     for slot, val in domain_bs.items() if slot not in blocked_slots]

            result = self.query_domain(domain, query)
            result = [Database.map_database_row(domain, k, domain_bs) for k in result]
            if return_results:
                all_results[domain] = (len(result), result)
            else:
                all_results[domain] = len(result)
        return all_results

    def save(self, path):
        shutil.copy(self.path, os.path.join(path, os.path.split(self.path)[-1]))
